@@ -47,7 +47,9 @@
 #include <sys/socket.h>
 #include <sys/un.h>
 #include <netinet/in.h>
+#ifdef HAVE_ARPA_INET_H
 #include <arpa/inet.h>
+#endif
 #include <sys/time.h>
 #include <sys/uio.h>
 #include <sys/kauth.h>
@@ -74,8 +76,8 @@
 #define VM_SYSTEM_HAS_GETGRNAM_R 1
 #endif
 
-# if defined(__FreeBSD__) && __FreeBSD_version >= 503000
-#  include <syslimits.h>  // PATH_MAX
+# if defined(__FreeBSD__)
+#  include <sys/syslimits.h>  // PATH_MAX
 # else
 #  include <limits.h>  // PATH_MAX
 # endif
@@ -179,8 +181,7 @@ Posix_Creat(ConstUnicode pathName, // IN:
  *      Open a file via POSIX fopen()
  *
  * Results:
- *      -1	Error
- *      >= 0	File descriptor (success)
+ *      A file pointer, or NULL on error.
  *
  * Side effects:
  *      errno is set on error
@@ -939,10 +940,15 @@ Posix_Execl(ConstUnicode pathName,   // IN:
    if (argv) {
       errno = 0;
       if (count > 0) {
-	 PosixConvertToCurrent(arg0, &argv[0]);
+         if (!PosixConvertToCurrent(arg0, &argv[0])) {
+            goto exit;
+         }
          va_start(vl, arg0);
          for (i = 1; i < count; i++) {
-	    PosixConvertToCurrent(va_arg(vl, char *), &argv[i]);
+            if (!PosixConvertToCurrent(va_arg(vl, char *), &argv[i])) {
+               va_end(vl);
+               goto exit;
+            }
          }
          va_end(vl);
       }
@@ -1011,16 +1017,21 @@ Posix_Execlp(ConstUnicode fileName,   // IN:
    if (argv) {
       errno = 0;
       if (count > 0) {
-	 PosixConvertToCurrent(arg0, &argv[0]);
+         if (!PosixConvertToCurrent(arg0, &argv[0])) {
+            goto exit;
+         }
          va_start(vl, arg0);
          for (i = 1; i < count; i++) {
-	    PosixConvertToCurrent(va_arg(vl, char *), &argv[i]);
+            if (!PosixConvertToCurrent(va_arg(vl, char *), &argv[i])) {
+               va_end(vl);
+               goto exit;
+            }
          }
          va_end(vl);
       }
       argv[count] = NULL;
       if (errno != 0) {
-	 goto exit;
+         goto exit;
       }
    }
 
@@ -2956,12 +2967,11 @@ exit:
  *
  * Posix_Printf --
  *
- *      Wrapper around printf. [w]printf can only print characters from
- *      the local character set.  So instead we use ConsoleWriteW which can
- *      print characters from the entire Unicode character set. 
+ *      POSIX printf.
  *
  * Returns:
- *      Returns the number of characters printed out.
+ *      Returns the number of characters printed out or a negative value on
+ *      failure.
  *
  * Side effects:
  *      None.
@@ -2998,10 +3008,11 @@ Posix_Printf(ConstUnicode format,
  *
  * Posix_Fprintf --
  *
- *      Wrapper around fprintf. 
- *      
+ *      POSIX fprintf.
+ *
  * Returns:
- *      Returns the number of bytes characters printed out.
+ *      Returns the number of characters printed out or a negative value on
+ *      failure.
  *
  * Side effects:
  *      None.
